# -*- coding: utf-8 -*-
from abc import ABCMeta
from abc import abstractmethod
from time import sleep
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from requests import Request
from requests import Session

from pyloniex.errors import PoloniexAPIError
from pyloniex.utils import RateLimiter


REQUESTS_PER_SECOND = 6


class PoloniexBaseAPI(metaclass=ABCMeta):

    @abstractmethod
    def __init__(
        self,
        *,
        requests_per_second: int = REQUESTS_PER_SECOND,
    ) -> None:
        self._rate_limiter = RateLimiter(
            requests_per_second,
            requests_per_second,
            1,
        )
        self._requests_per_second = requests_per_second
        self._session = Session()

    def request(
        self,
        *args,
        **kwargs,
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        cls_name = type(self).__name__
        while self._rate_limiter.check(cls_name) is False:
            sleep(1 / self._requests_per_second)

        request = Request(*args, **kwargs)
        prepared = self._session.prepare_request(request)

        response = self._session.send(prepared)
        data = response.json()
        if 'error' in data:
            raise PoloniexAPIError(response.status_code, data['error'])

        return data



# -*- coding: utf-8 -*-
from collections import OrderedDict
from time import time
from typing import Tuple


class RateLimiter:
    """A very fast rate in-memory rate limiter using the token bucket
    algorithm. See https://en.wikipedia.org/wiki/Token_bucket for details.

    Tokens accumulate in the bucket at a constant rate per second, given by
    `per_second`. Each time RateLimiter::check is called, one token is removed
    from the bucket.

    The `burst` parameter controls the maximum number of tokens that may be
    held by the bucket at once. If tokens are not removed, they will accumulate
    up to this limit.

    Events are identified by a key (so a single RateLimiter instance can limit
    multiple kinds of events, or events from multiple sources, etc.). `size`
    limits the number of keys that can be tracked. If this limit is reached,
    the oldest key will be evicted.

    If you just want to limit some event to no more than X times per second,
    set `per_second` = `burst` = X.

    If unused capacity can be "banked" for increased usage later, set `burst`
    to some value > `per_second`.
    """

    def __init__(self, per_second: float, burst: float, size: int) -> None:
        self._per_second = per_second
        self._burst = burst
        self._size = size
        self._buckets: OrderedDict[str, Tuple[float, float]] = OrderedDict()

    def check(self, key: str) -> bool:
        now = time()
        budget, timestamp = self._buckets.get(key, (self._burst, now))
        budget = min(self._burst, budget + (now - timestamp) * self._per_second) - 1
        if budget < 0:
            return False

        if key in self._buckets:
            del self._buckets[key]
        elif len(self._buckets) >= self._size:
            self._buckets.popitem(last=False)

        self._buckets[key] = (budget, now)
        return True



# -*- coding: utf-8 -*-


class PoloniexAPIError(Exception):

    def __init__(self, status_code: int, message: str) -> None:
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return f'[{self.status_code}] {self.message}'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup


setup(
    name='pyloniex',
    version='0.0.1',

    packages=find_packages(),

    install_requires=[
        'cryptography',
        'requests',
    ],

    author='Adam Rothman',
    author_email='rothman.adam@gmail.com',
    description='Python bindings for the Poloniex API',
    url='https://github.com/adamrothman/pyloniex',
)
